﻿# Author: Anthony Stahler
# Last updated: Sept 27, 2022
# License: Free for personal use, for commercial use please reach out to the author via the BestEdit contact page: http://bestedit.net/contact/
# The most recent version of this script can be found at http://bestedit.net
#
# This script takes an xml fanedit file and one or more srt (subtitle) files and outputs a companion srt file to match the xml fanedit file.
# It expects a 24fps xml file generated by the premiere-pro-xml-fix script - this is important because there are potential bugs and issues with 
# the Premiere XML calculations, even if it's saved out at 24fps, and this script expects the data to be correct.
#
# IMPORTANT: make sure you've processed your xml files with $cleanvidid = $true if you're subtitling multiple xml at once, otherwise you may have conflicting ids that will confuse the code
#
# This script assumes your source files are mp4, if not you need to change $videoext = ".mp4" to reflect the file type you use
#

# SRT format info:
# Each subtitle has four parts in the SRT file.
# -A numeric counter indicating the number or position of the subtitle.
# -Start and end time of the subtitle separated by --> characters
# hours:minutes:seconds,milliseconds (00:00:00,000) format
# -Subtitle text in one or more lines.
# -A blank line indicating the end of the subtitle.

function ConvertSrtTimeToFrames
{
    param($timestring)

    [decimal]$totalmillis = -1

    $millisarr = $timestring.Trim().Split(',')

    if ($millisarr.Length -eq 2)
    {
        $totalmillis = [System.Convert]::ToInt64($millisarr[1])
        
        $millisarr = $millisarr[0].Split(':')

        if ($millisarr.Length -eq 3)
        {
            $totalmillis += ([System.Convert]::ToInt64($millisarr[2]) * 1000) # seconds
            $totalmillis += ([System.Convert]::ToInt64($millisarr[1]) * 60000) # minutes
            $totalmillis += ([System.Convert]::ToInt64($millisarr[0]) * 3600000) # hours

            # divide millisecond timestamp by 24fps (1000/24) = 41.66666666666667
            $totalmillis = $totalmillis / (1000.0/24.0)
        }
        else
        {
            # not right format, clear it
            $totalmillis = -1
        }
    }

    #string to long
    #$framestring = [System.Convert]::ToInt64($framestring)

    #fix the math (could be 1 frame wrong depending on the initial erroneous number that was rounded)
    # note that the subtraction is the opposite of doing this with framerates, since ticks go up as framerate goes down
    #$framestring = [math]::Round($framestring + ((($fileticksperframe-$pproconst)/$pproconst)*$framestring))

    return [decimal]$totalmillis
}

function ConvertFramesToSrtTime
{
    param($totalframes)

    $timestring = ""
    $totalframes = [decimal]$totalframes # important, things get wonky if you don't explicitly cast this - the following multiply concatenates rather than multiplies.

    # multiply frames by 24fps (1000/24) = 41.66666666666667 to get milliseconds
    [decimal]$totalmillis = $totalframes * (1000.0/24.0)

    # round down, we'll remove the remainder in a second
    $tmpmillis = [math]::Floor($totalmillis / 3600000)
    $timestring += $tmpmillis.toString("00") + ":"
    $totalmillis -= ($tmpmillis * 3600000)

    $tmpmillis = [math]::Floor($totalmillis / 60000)
    $timestring += $tmpmillis.toString("00") + ":"
    $totalmillis -= ($tmpmillis * 60000)

    $tmpmillis = [math]::Floor($totalmillis / 1000)
    $timestring += $tmpmillis.toString("00") + ","
    $totalmillis -= ($tmpmillis * 1000)

    # add the last milliseconds on the end
    $timestring += $totalmillis.toString("000")

    return [string]$timestring
}


# --------------------------
# SETTINGS
# --------------------------

$srcpath = "M:\movie-edits-exports\auto-subtitle\"

$shortenthreshold = 0.70 # anything 70% or above will be automatically fit to the clip so that you don't get overlapping subtitles, anything less than the threshold you need to handle manually

$targetaudiochannel = 0 # theoretically your audio channels should all be the same, but if they're not you can change this to target your desired channel

# --------------------------
# --------------------------

$speedenhancement = $true 

$videoext = ".mp4"
$outputpath = $srcpath + "sub_output\"
$targetext = "*.xml" 
$targetext2 = "*.srt" 

$masterclip_ht = @{}
$subtitles_ht = @{}

# constant for 24fps 
[long]$pproconst = 10584000000

# make output path if it doesn't exist
$silenceoutput = New-Item -ItemType Directory -Force -Path $outputpath

Write-Output "Extracting source videos from xml files..."

# start extracting name/id pairs from xml files
# this is a pre-processing step, we'll come back and subtitle each xml later after we cache the SRT
Get-ChildItem -path ( $srcpath + $targetext ) | foreach {
    $origname1 = $_.Name
    $origname1NOext = $_.BaseName
    $origname1_normal = $origname1 -replace '[\[\]\(\)_,]', ' ' # stage 1, replace crap with spaces
    $origname1_normal = $origname1_normal -replace '\s+', '.' # stage 2, replace multiple spaces in a row with one period
    $ismatch = $origname1_normal -match '(.+)(1080p|720p)'

    if ($ismatch -eq $true)
    {
        $isproject = $false
        [long]$masterclip_id = -1
        
        $tmpstr = ""

        [System.IO.File]::ReadLines($srcpath + $origname1) | ForEach-Object {

                $linedata_orig = $_

                switch -regex ($linedata_orig)
                {
                    '(<clipitem )' { 
                        $masterclip_id = -1
                        break
                    }

                    '(</clipitem>)' { 
                        $masterclip_id = -1
                        break
                    }

                    '(<masterclipid>)' { 
                        $tmpstr = $linedata_orig -replace '(<masterclipid>)(.*)(</masterclipid>)', "`$2"
                        $tmpstr = $tmpstr.ToLower().Trim()

                        # make the result just the number
                        $masterclip_id = $tmpstr.Replace("masterclip-", "")
                        break
                    }

                    # found a <name> tag
                    '(<name>)(.*)(</name>)' { 
                        if ($masterclip_id -ne -1)
                        {
                            $tmpstr = $linedata_orig -replace '(<name>)(.*)(</name>)', "`$2"
                            $tmpstr = $tmpstr.Trim()

                            if ($masterclip_ht.ContainsKey($tmpstr) -eq $false)
                            {
                                $masterclip_ht.add($tmpstr, $masterclip_id)
                                #$tmpstr = "file: " + $tmppair.Value + ", index: " + $tmppair.Name
                            }

                            # don't want to accidentally pick up a name from a filter or something
                            $masterclip_id = -1
                        }
                        break
                    }

                    '(<project>)' { 
                        # we only want sequences, not projects, warn the user
                        $isproject = $true
                        break
                    }

                }
        }

        if ($isproject -eq $true)
        {
            Write-Output "Project detected, results may not be accurate for file: '$origname1'"
        }

    }
    else
    {
        Write-Output "File '$origname1' not in normal movie format (Name.Year.Resolution) or tv format (Name.S##E##.Resolution), skipping!"
    }
}

Write-Output "Loading all srt files..."

# load our srt database
Get-ChildItem -path ( $srcpath + $targetext2 ) | foreach {
    $srtfilename = $_.Name
    $srtfileasmp4 = $_.BaseName + $videoext

    $srtfileindex = $masterclip_ht[$srtfileasmp4]

    # don't bother processing if we can't find an index
    if ($srtfileindex -ne $null)
    {
        $insubtitle = $false
        [decimal]$timestamp_start = -1
        [decimal]$timestamp_end = -1
        $subtitle_text = ""
        $sub_arr = @()


        [System.IO.File]::ReadLines($srcpath + $srtfilename) | ForEach-Object {
            $linedata = $_

            $linedata = $linedata.Trim()

            if ($linedata.Length -eq 0)
            {
                if ($insubtitle -eq $true)
                {
                    # reached the end of the current subtitle
                    $insubtitle = $false

                    if ($timestamp_start -gt -1 -and $timestamp_end -gt -1 -and $subtitle_text.Length -gt 0)
                    {
                        # we could create this in the loop but it's cleaner just to shove it in at the end
                        $newsub = @{}

                        # data to save to our temporary srt hashtable
                        $newsub.TimeStart = $timestamp_start
                        $newsub.TimeEnd = $timestamp_end
                        $newsub.Subtext = $subtitle_text

                        # shove the hashtable into the array, which we'll later put in another hashtable for easy lookup
                        $sub_arr += $newsub
                    }

                    # clear data
                    $timestamp_start = -1
                    $timestamp_end = -1
                    $subtitle_text = ""
                }
            }
            else
            {
                if ($timestamp_start -eq -1 -and $timestamp_end -eq -1 -and $linedata.IndexOf('-->') -gt 0 -and $insubtitle -eq $false)
                {
                    #$tmparr = $linedata.Split('-->') # doesn't work, treats string as array of characters and gives us 4 length array rather than 2 as expected
                    $tmparr = $linedata -split '-->'

                    if ($tmparr.Length -eq 2)
                    {
                        $timestamp_start = ConvertSrtTimeToFrames($tmparr[0])
                        $timestamp_end = ConvertSrtTimeToFrames($tmparr[1])
                    }

                    $insubtitle = $true
                }   
                elseif ($insubtitle -eq $true)
                {
                    $subtitle_text += "$linedata`n"
                }
            }
        }

        # add the subtitles to a hashtable using the video index so we can pull them easily later as we process the xml
        $subtitles_ht.add("$srtfileindex", $sub_arr)
    }
    else
    {
        # we probably don't need a warning about this, just means we have more subs than we need
        # later there's another warning for videos that it can't find a subtitle to go along with, so if the user has them named incorrectly they'll get warned at that point
        #Write-Output "Can't find an associated video index for: '$srtfilename', make sure your subtitles have exactly the same name as the original video they're associated with (e.g. if you used my.video.1999.1080p.mp4 in your fanedit you should have a subtitle file named my.video.1999.1080p.srt)"
    }

}

# check if we have subtitles for all the expected video sources, if not then warn the user
foreach ($tmppair in $masterclip_ht.GetEnumerator())
{
    # make sure we're using the key as a string to avoid conversion issues
    $tmpstr = $tmppair.Value
    $sub_arr = $subtitles_ht["$tmpstr"]

    if ($sub_arr -eq $null)
    {
        $tmpstr = $tmppair.Name
        Write-Output "WARNING: Missing subtitle '$tmpstr', make sure your subtitles have exactly the same name as the original video they're associated with `n`t`t(e.g. if you used my.video.1999.1080p.mp4 in your fanedit you should have a subtitle file named my.video.1999.1080p.srt)`n`t`tIf '$tmpstr' is not the name you expected, double check that you set `$cleanfilenames = `$false in the premiere-pro-xml-fix script so it doesn't normalize your filenames`n"
    }
}

Write-Output "Processing xml files against subtitle database..."

if ($subtitles_ht.Count -gt 0)
{
    # now that we have the XMLs preprocessed and the SRTs indexed it's time to actually process the subtitles
    Get-ChildItem -path ( $srcpath + $targetext ) | foreach {
        $origname1 = $_.Name
        $origname1NOext = $_.BaseName

        Write-Output "Processing '$origname1'"

        $primary_audio_track = $false
        $masterclip_id = -1
        $currentsubarr = $null

        # explicitly typing these is super important, powershell is terrible at knowing when to cast or concatenate
        [decimal]$clipspeed = -1
        [long]$clipframein = 0
        [long]$clipframeout = 0
        [long]$clipstart = -1
        [long]$clipend = -1
        [long]$clipin = -1
        [long]$clipout = -1

        Write-Output "`t`tReading and standardizing format..."
        # grab the entire file so we can pre-process it in certain ways before doing the conversion pass
        $wholefile = [System.IO.File]::ReadAllText($srcpath + $origname1)

        # standardize newlines
        $wholefile = $wholefile -replace "`r`n", "`n"
        $wholefile = $wholefile -replace "`r", "`n"

        # quick loop through to drop any tracks that aren't necessary for our subtitles, so we don't spend time regexing them
        if ($speedenhancement -eq $true)
        {
            $wholefile_lines = $wholefile -split "`n"
            $wholefile = ""
            $cullline = $false

            ForEach ($linedata_orig in $wholefile_lines) {
                switch -regex ($linedata_orig)
                {
                    '(<track )(.*)(currentExplodedTrackIndex)(.*)(totalExplodedTrackCount)' { 
                        # audio track, see if it's one we need
                        $tmplng = ($linedata_orig -replace '(?ms)(.*currentExplodedTrackIndex=")(.*?)(".*)', "`$2").Trim()

                        if ($tmplng -ne $targetaudiochannel)
                        {
                            # not our $primary_audio_track, drop it
                            $cullline = $true
                        }
                        break
                    }
                    '(<track )' { 
                        # non-audio track, drop it
                        $cullline = $true
                        break
                    }
                    '(</track>)' { 
                        if ($cullline -eq $true)
                        {
                            $linedata_orig = "" # drop end track tag before we turn cull off
                        }

                        $cullline = $false
                        break
                    }

                }

                if ($cullline -eq $false)
                {
                    $wholefile += "$linedata_orig`n"
                }
            }

            #$wholefile = $wholefile_lines -join "`n"
        }

        Write-Output "`t`tRegex preprocessing stage 1..."
        # The speed effect is hard to fix in the original format, we're going to regex it to make some changes and insert our own tags that make it easier to reprocess accurately
        while ($wholefile -match '(?ms)(<clipitem ((?!<clipitem ).)*?<effectid>timeremap</effectid>.*?</clipitem>)')
        {
            $speedclip = $wholefile -replace '(?ms)(.*)(<clipitem ((?!<clipitem ).)*?<effectid>timeremap</effectid>.*?</clipitem>)(.*)', "`$2"
            $speedclip_orig = $speedclip

            # grab the multiplier for the speed
            #
            #<parameter authoringApp="PremierePro">
		    #	<parameterid>speed</parameterid>
		    #	<name>speed</name>
		    #	<valuemin>-100000</valuemin>
		    #	<valuemax>100000</valuemax>
		    #	<value>92.3592</value>
		    #</parameter>
            $speedparam = $speedclip -replace '(?ms)(.*)(<parameter ((?!<parameter ).)*?<parameterid>speed</parameterid>.*?<value>(.*?)</value>.*?</parameter>)(.*)', "`$4"
                    
            # Now attach the speed as a new tag right after the start of the clip, so we have it when we're processing the other variables
            # <clipitem id="clipitem-329">
            $speedclip = $speedclip -replace '(?ms)(<clipitem ((?!>).)*?>)', "`$1`n`t`t`t`t`t`t<besteditspeed>$speedparam</besteditspeed>"

            # get rid of this so it doesn't match again
            $speedclip = $speedclip.Replace('<effectid>timeremap</effectid>','')

            # put our reprocessed speed clip back into the place of the original
            $wholefile = $wholefile.Replace($speedclip_orig, $speedclip)
        }

        Write-Output "`t`tRegex preprocessing stage 2 (this may take a while, be patient)..."
        # We want to convert the transitions to tags within the clip XML to make it easier to process each clip without having to look at data outside it

        # This regex 0 is matching one clip, two transitions, and one clip, and we don't want it to match in that situation, only with one transition in the middle
        # the first wildcard is eating the entire first transition, forbid that
        # '(?ms)(<clipitem ((?!<clipitem ).)*?<effectid>KGAudioTransCrossFade0dB</effectid>((?!<transitionitem>).)*?</clipitem>)'
        # once we fix that one the second also starts eating double transitions, first it does one then the next loop it grabs the second one, apply the same fix to it
        # add a preemptive fix to the last regex too so it doesn't grab multiple clips in an edge case

        $fade_regex = @('(?ms)(<clipitem ((?!<clipitem )(?!</transitionitem>).)*?<effectid>(KGAudioTransCrossFade0dB|KGAudioTransCrossFade3dB|KGAudioTransCrossFadeXdB)</effectid>((?!<transitionitem>).)*?</clipitem>)', # this regex should match one clip, one transition, and another clip
                        '(?ms)(<clipitem ((?!<clipitem )(?!</transitionitem>).)*?<effectid>(KGAudioTransCrossFade0dB|KGAudioTransCrossFade3dB|KGAudioTransCrossFadeXdB)</effectid>((?!<transitionitem>).)*?</transitionitem>)', # this regex should match one clip, one transition
                        '(?ms)(<transitionitem>((?!</transitionitem>).)*?<effectid>(KGAudioTransCrossFade0dB|KGAudioTransCrossFade3dB|KGAudioTransCrossFadeXdB)</effectid>((?!<transitionitem>)(?!</clipitem>).)*?</clipitem>)') # this regex should match one transition, one clip

        $fade_regex2 = @('(?ms)(.*)(<clipitem ((?!<clipitem )(?!</transitionitem>).)*?<effectid>(KGAudioTransCrossFade0dB|KGAudioTransCrossFade3dB|KGAudioTransCrossFadeXdB)</effectid>((?!<transitionitem>).)*?</clipitem>)(.*)',
                        '(?ms)(.*)(<clipitem ((?!<clipitem )(?!</transitionitem>).)*?<effectid>(KGAudioTransCrossFade0dB|KGAudioTransCrossFade3dB|KGAudioTransCrossFadeXdB)</effectid>((?!<transitionitem>).)*?</transitionitem>)(.*)',
                        '(?ms)(.*)(<transitionitem>((?!</transitionitem>).)*?<effectid>(KGAudioTransCrossFade0dB|KGAudioTransCrossFade3dB|KGAudioTransCrossFadeXdB)</effectid>((?!<transitionitem>)(?!</clipitem>).)*?</clipitem>)(.*)')

        $currenttime = [math]::Floor( (New-TimeSpan -Start (Get-Date "01/01/1970") -End (Get-Date)).TotalSeconds )

        for ($i=0;$i -lt $fade_regex.Length;$i++)
        {
            while ($wholefile -match $fade_regex[$i])
            {
                $speedclip = $wholefile -replace $fade_regex2[$i], "`$2"
                $speedclip_orig = $speedclip

                $tmpcenter = $false
                [long]$tmplng = 0
                [long]$tmplng2 = 0
                [long]$tmplng3 = 0
                [long]$tmplng4 = 0
                [long]$tmplng5 = 0

                # for our first regex where we get two clips with a transition in the center we want to check if both are attached to the transition
                # the value -1 in start/end indicates the clip is attached to the transition it's next to
                if ($i -eq 0)
                {
                    # get END from first clip, and START from second clip
                    $tmplng = ($speedclip -replace '(?ms)(.*)(<clipitem )(.*?)<end>(.*?)</end>(.*?)(<transitionitem>)(.*)', "`$4").Trim()
                    $tmplng2 = ($speedclip -replace '(?ms)(.*)(</transitionitem>)(.*?)<start>(.*?)</start>(.*)', "`$4").Trim()

                    if ($tmplng -eq -1 -and $tmplng2 -eq -1)
                    {
                        $tmpcenter = $true
                    }
                }

                # grab the alignment
                $speedparam = $speedclip -replace '(?ms)(.*)(<transitionitem>((?!</transitionitem>).)*?<alignment>(.*?)</alignment>.*?</transitionitem>)(.*)', "`$4"
                $speedparam = $speedparam.ToLower().Trim()

                # grab the ticks, we're going to use these in our new tag to simplify things later
                $tmplng = ($speedclip -replace '(?ms)(.*)(<transitionitem>((?!</transitionitem>).)*?<pproTicksIn>(.*?)</pproTicksIn>.*?</transitionitem>)(.*)', "`$4").Trim()
                $tmplng2 = ($speedclip -replace '(?ms)(.*)(<transitionitem>((?!</transitionitem>).)*?<pproTicksOut>(.*?)</pproTicksOut>.*?</transitionitem>)(.*)', "`$4").Trim()
                $tmplng3 = ($speedclip -replace '(?ms)(.*)(<transitionitem>((?!</transitionitem>).)*?<cutPointTicks>(.*?)</cutPointTicks>.*?</transitionitem>)(.*)', "`$4").Trim()

                # use the cutpointticks to figure out how the transition is shifted
                $tmplng4 = $tmplng3 # fade out
                $tmplng5 = $tmplng3 # fade in

                # transition length in ticks
                $tmplng3 = $tmplng2 - $tmplng

                # use the cutpointticks to figure out how the transition is shifted
                # start is shifted to the right, typically with 0 for cutpointticks
                # end is shifted to the left, typically with the full length for cutpointticks
                # there may be small variations for rounding and bypassing premiere bugs
                # examples:
                # 20 frame transition, middle, both values should be 10 (approx, might be a little different depending on if it's offset slightly)
                # 20 frame transition, start, fade in should be 0, fade out should be 20
                # 20 frame transition, end, fade in should be 20, fade out should be 0
                # start and end may seem like they're backward, it's counterintuitive because for a left-shifted transition (end), there is no extra length added to the end of the first clip since the transition is fully overlapping it
                # the full length of the transition IS added to the start of the second clip so that it's long enough to overlap the previous one underneath the transition
                $tmplng4 = $tmplng3 - $tmplng4 # fade out, length - cutpointticks
                $tmplng5 = $tmplng3 - $tmplng4 # fade in, length - part used by fade out

                # convert to frames
                $tmplng3 = $tmplng3 / $pproconst
                $tmplng = $tmplng / $pproconst
                $tmplng2 = $tmplng2 / $pproconst
                $tmplng4 = $tmplng4 / $pproconst
                $tmplng5 = $tmplng5 / $pproconst

                # adjust our clip start/end by part of the transition length
                $tmplng = $tmplng + $tmplng3
                $tmplng2 = $tmplng2 - $tmplng3



                # Now attach the indicator tags right after the start of each clip, so we have it when we're processing the other variables
                # We need to attach two different values because the OUT-IN length of a clip needs some subtracted off each side to accurately frame the subtitles, since the clip extends itself the full length of the transition on both sides
                # So we'll deliver a tag with the amount to subtract on each side, as well as the true START/END values for a clip since they're -1 internally if a transition is applied
                # <clipitem id="clipitem-329">
                switch ($speedparam)
                {
                    'center' {
                        $speedclip = $speedclip -replace '(?ms)(<clipitem ((?!>).)*?>)(.*)(<clipitem ((?!>).)*?>)', "`$1`n`t`t`t`t`t`t<besteditfadeout>$tmplng4</besteditfadeout>`n`t`t`t`t`t`t<besteditclipend>$tmplng</besteditclipend>`$3`$4`n`t`t`t`t`t`t<besteditfadein>$tmplng5</besteditfadein>`n`t`t`t`t`t`t<besteditclipstart>$tmplng2</besteditclipstart>"
                        break
                    }

                    'end-black' {
                        if ($tmpcenter -eq $true)
                        {
                            # this should never be true, but just in case...
                            Write-Output "Warning: End-Black conversion anomaly at '$tmplng'"
                            $speedclip = $speedclip -replace '(?ms)(<clipitem ((?!>).)*?>)(.*)(<clipitem ((?!>).)*?>)', "`$1`n`t`t`t`t`t`t<besteditfadeout>$tmplng4</besteditfadeout>`n`t`t`t`t`t`t<besteditclipend>$tmplng</besteditclipend>`$3`$4`n`t`t`t`t`t`t<besteditfadein>$tmplng5</besteditfadein>`n`t`t`t`t`t`t<besteditclipstart>$tmplng2</besteditclipstart>"
                        }
                        else
                        {
                            $speedclip = $speedclip -replace '(?ms)(<clipitem ((?!>).)*?>)(.*)(<transitionitem>)', "`$1`n`t`t`t`t`t`t<besteditfadeout>$tmplng4</besteditfadeout>`n`t`t`t`t`t`t<besteditclipend>$tmplng</besteditclipend>`$3`$4"
                        }
                        break
                    }

                    'end' {
                        if ($tmpcenter -eq $true)
                        {
                            $speedclip = $speedclip -replace '(?ms)(<clipitem ((?!>).)*?>)(.*)(<clipitem ((?!>).)*?>)', "`$1`n`t`t`t`t`t`t<besteditfadeout>$tmplng4</besteditfadeout>`n`t`t`t`t`t`t<besteditclipend>$tmplng</besteditclipend>`$3`$4`n`t`t`t`t`t`t<besteditfadein>$tmplng5</besteditfadein>`n`t`t`t`t`t`t<besteditclipstart>$tmplng2</besteditclipstart>"
                        }
                        else
                        {
                            $speedclip = $speedclip -replace '(?ms)(<clipitem ((?!>).)*?>)(.*)(<transitionitem>)', "`$1`n`t`t`t`t`t`t<besteditfadeout>$tmplng4</besteditfadeout>`n`t`t`t`t`t`t<besteditclipend>$tmplng</besteditclipend>`$3`$4"
                        }
                        break
                    }

                    'start-black' {
                        if ($tmpcenter -eq $true)
                        {
                            # this should never be true, but just in case...
                            Write-Output "Warning: Start-Black conversion anomaly at '$tmplng'"
                            $speedclip = $speedclip -replace '(?ms)(<clipitem ((?!>).)*?>)(.*)(<clipitem ((?!>).)*?>)', "`$1`n`t`t`t`t`t`t<besteditfadeout>$tmplng4</besteditfadeout>`n`t`t`t`t`t`t<besteditclipend>$tmplng</besteditclipend>`$3`$4`n`t`t`t`t`t`t<besteditfadein>$tmplng5</besteditfadein>`n`t`t`t`t`t`t<besteditclipstart>$tmplng2</besteditclipstart>"
                        }
                        else
                        {
                            $speedclip = $speedclip -replace '(?ms)(</transitionitem>)(.*)(<clipitem ((?!>).)*?>)', "`$1`$2`$3`n`t`t`t`t`t`t<besteditfadein>$tmplng5</besteditfadein>`n`t`t`t`t`t`t<besteditclipstart>$tmplng2</besteditclipstart>"
                        }
                        break
                    }

                    'start' {
                        if ($tmpcenter -eq $true)
                        {
                            $speedclip = $speedclip -replace '(?ms)(<clipitem ((?!>).)*?>)(.*)(<clipitem ((?!>).)*?>)', "`$1`n`t`t`t`t`t`t<besteditfadeout>$tmplng4</besteditfadeout>`n`t`t`t`t`t`t<besteditclipend>$tmplng</besteditclipend>`$3`$4`n`t`t`t`t`t`t<besteditfadein>$tmplng5</besteditfadein>`n`t`t`t`t`t`t<besteditclipstart>$tmplng2</besteditclipstart>"
                        }
                        else
                        {
                            $speedclip = $speedclip -replace '(?ms)(</transitionitem>)(.*)(<clipitem ((?!>).)*?>)', "`$1`$2`$3`n`t`t`t`t`t`t<besteditfadein>$tmplng5</besteditfadein>`n`t`t`t`t`t`t<besteditclipstart>$tmplng2</besteditclipstart>"
                        }
                        break
                    }

                }

                # wipe out the effectid tag so we don't match this one again
                $speedclip = $speedclip -replace '(?ms)(<effectid>[^<]*</effectid>)', ''

                # put our reprocessed clip back into the place of the original
                #$wholefile = $wholefile -replace $speedclip_orig, $speedclip
                $wholefile = $wholefile.Replace($speedclip_orig, $speedclip)
            }
        }

        Write-Output "`t`tData processing..."
        Write-Output "`t`tThreshold set to '$shortenthreshold', any overlap pct equal or greater than this will be automatically shortened to fit"
        

        # now that our helper tags are in place we can easily process in a single sequential loop through the file
        $wholefile_lines = $wholefile -split "`n"

        # the IN/OUT tags are positions in each of the source video files
        # the START/END tags are positions in the timeline/sequence, if a clip has a transition the START/END will be -1 and the actual value will be stored in the transition adjacent to the clip
        # the OUT-IN length of a clip is equal to the length of the clip PLUS the full length of the transition on either side.
        $output_arr = @()

        ForEach ($linedata_orig in $wholefile_lines) {

            switch -regex ($linedata_orig)
            {
                '(<track )(.*)(currentExplodedTrackIndex)(.*)(totalExplodedTrackCount)' { 
                    # need to ignore video clips, only want the first channel of each audio track

                    # audio track is going to look something like this originally:
                    # <track hardpan="TRUE" TL.SQTrackAudioKeyframeStyle="0" TL.SQTrackShy="0" TL.SQTrackExpandedHeight="25" TL.SQTrackExpanded="0" MZ.TrackTargeted="1" currentExplodedTrackIndex="0" totalExplodedTrackCount="6" premiereTrackType="5.1">
                    # stereo:
                    # <track TL.SQTrackAudioKeyframeStyle="0" TL.SQTrackShy="0" TL.SQTrackExpandedHeight="25" TL.SQTrackExpanded="0" MZ.TrackTargeted="1" PannerCurrentValue="0.5" PannerIsInverted="true" PannerStartKeyframe="-91445760000000000,0.5,0,0,0,0,0,0" PannerName="Balance" currentExplodedTrackIndex="0" totalExplodedTrackCount="2" premiereTrackType="Stereo">
                    # mono:
                    # <track TL.SQTrackAudioKeyframeStyle="0" TL.SQTrackShy="0" TL.SQTrackExpandedHeight="25" TL.SQTrackExpanded="0" MZ.TrackTargeted="1" PannerCurrentValue="0.5" PannerIsInverted="true" PannerStartKeyframe="-91445760000000000,0.5,0,0,0,0,0,0" PannerName="Balance" currentExplodedTrackIndex="0" totalExplodedTrackCount="1" premiereTrackType="Stereo">

                    # Now we want to check how many tracks we have, and which track we're on
                    # use ? so it isn't greedy, important here
                    $tmplng = ($linedata_orig -replace '(?ms)(.*currentExplodedTrackIndex=")(.*?)(".*)', "`$2").Trim()

                    # track 0, first one
                    if ($tmplng -eq $targetaudiochannel)
                    {
                        $primary_audio_track = $true
                    }
                    break
                }

                '(</track>)' { 
                    $primary_audio_track = $false
                    $masterclip_id = -1
                    $clipspeed = -1
                    $clipframein = 0
                    $clipframeout = 0
                    $clipstart = -1
                    $clipend = -1
                    $clipin = -1
                    $clipout = -1
                    break
                }

                '(<clipitem )' { 
                    $masterclip_id = -1
                    $clipspeed = -1
                    $clipframein = 0
                    $clipframeout = 0
                    $clipstart = -1
                    $clipend = -1
                    $clipin = -1
                    $clipout = -1
                    break
                }

                '(</clipitem>)' { 
                    $masterclip_id = -1
                    $clipspeed = -1
                    $clipframein = 0
                    $clipframeout = 0
                    $clipstart = -1
                    $clipend = -1
                    $clipin = -1
                    $clipout = -1
                    break
                }

                '(<transitionitem>)' { 
                    $masterclip_id = -1
                    $clipspeed = -1
                    $clipframein = 0
                    $clipframeout = 0
                    $clipstart = -1
                    $clipend = -1
                    $clipin = -1
                    $clipout = -1
                    break
                }

                '(<masterclipid>)' { 
                    $tmpstr = $linedata_orig -replace '(<masterclipid>)(.*)(</masterclipid>)', "`$2"
                    $tmpstr = $tmpstr.ToLower().Trim()

                    # make the result just the number
                    $masterclip_id = $tmpstr.Replace("masterclip-", "")

                    $currentsubarr = $subtitles_ht["$masterclip_id"]
                    break
                }

                '(<besteditspeed>)(.*)(</besteditspeed>)' { 
                    $clipspeed = [System.Convert]::ToDecimal(($linedata_orig -replace '(<besteditspeed>)(.*)(</besteditspeed>)', "`$2").Trim())
                    # convert to percent, the original tag has it as a whole number
                    $clipspeed = $clipspeed / 100
                    break
                }

                '(<besteditfadein>)(.*)(</besteditfadein>)' { 
                    # found one of our cached values, add these frames to the start of the clip
                    $clipframein = ($linedata_orig -replace '(<besteditfadein>)(.*)(</besteditfadein>)', "`$2").Trim()
                    break
                }

                '(<besteditfadeout>)(.*)(</besteditfadeout>)' { 
                    # found one of our cached values, subtract these frames from the end of the clip
                    $clipframeout = ($linedata_orig -replace '(<besteditfadeout>)(.*)(</besteditfadeout>)', "`$2").Trim()
                    break
                }

                '(<besteditclipstart>)(.*)(</besteditclipstart>)' { 
                    # found one of our cached values, subtract these frames from the end of the clip
                    $clipstart = ($linedata_orig -replace '(<besteditclipstart>)(.*)(</besteditclipstart>)', "`$2").Trim()
                    break
                }

                '(<besteditclipend>)(.*)(</besteditclipend>)' { 
                    # found one of our cached values, subtract these frames from the end of the clip
                    $clipend = ($linedata_orig -replace '(<besteditclipend>)(.*)(</besteditclipend>)', "`$2").Trim()
                    break
                }

                '(<in>)(.*)(</in>)' { 
                    $clipin = ($linedata_orig -replace '(<in>)(.*)(</in>)', "`$2").Trim()
                    break
                }

                '(<out>)(.*)(</out>)' { 
                    $clipout = ($linedata_orig -replace '(<out>)(.*)(</out>)', "`$2").Trim()
                    break
                }

                '(<start>)(.*)(</start>)' { 
                    # only want to assign if we don't already have a value from our cached transition tags
                    if ($clipstart -eq -1)
                    {
                        $clipstart = ($linedata_orig -replace '(<start>)(.*)(</start>)', "`$2").Trim()
                    }
                    break
                }

                '(<end>)(.*)(</end>)' { 
                    # only want to assign if we don't already have a value from our cached transition tags
                    if ($clipend -eq -1)
                    {
                        $clipend = ($linedata_orig -replace '(<end>)(.*)(</end>)', "`$2").Trim()
                    }
                    break
                }

            }

            if ($primary_audio_track -eq $true)
            {
                # if we aren't processing the first channel of an audio track we can safely ignore the incoming data

                # at minimum we need to know what subtitle file to associate, the start/end and in/out values so we can search and correctly position the subtitles. Speed, framein/frameout are optional and used if they're set
                if ($masterclip_id -ne -1 -and $clipstart -ne -1 -and $clipend -ne -1 -and $clipin -ne -1 -and $clipout -ne -1)
                {
                    # process subtitles
                    if ($currentsubarr -is [array]) # for some reason checking against null here returns true if the array is empty, check data type instead
                    {
                        if ($clipspeed -ne -1)
                        {
                            # this fixes the in/out durations, which are drastically wrong when a speed modifier is added to the clip
                            # not sure why they use completely wrong integers rather than using decimals like rational programmers
                            $clipin = $clipin * $clipspeed
                            $clipout = $clipout * $clipspeed

                            # WARNING: There's actually an obscure bug I'm not fixing here, if anyone actually uses speed modifiers on subtitles they'll probably get more subs than they want
                            # this is due to the in/out values being incorrectly preprocessed by the xml-fix script, so for a 10% speed clip the in/out duration is actually 
                            # still 10 times longer than it should be even after removing the clipspeed modifier (in-out=24 instead of 2.4 for a 1 second clip)
                            # not fixing this bug now since it's unlikely people will use audio speed modifiers much and the in/out being longer doesn't matter for importing xml edits
                            Write-Output "WARNING: Audio clip with modified duration detected, this is not fully implemented and may result in extra subtitles for the clip"

                            # get rid of speed, don't need it anymore
                            $clipspeed = -1
                        }


                        # we want to subtract half the transitions for our comparison to simulate the not very audible half of the transitions, that way we don't subtitle something the user couldn't hear
                        $clipframein = $clipin + $clipframein
                        $clipframeout = $clipout - $clipframeout

                        #$testarr = @()
                        
                        for ($i=0;$i -lt $currentsubarr.Count;$i++)
                        {
                            $savesub = $false
                            $subwarning = ""
                            $thresholdstart = 0.0
                            $thresholdend = 0.0

                            if ($clipframein -le $currentsubarr[$i].TimeStart -and $clipframeout -ge $currentsubarr[$i].TimeEnd)
                            {
                                # subtitle is fully within the bounds of this clip, keep it
                                # CI < S > CO
                                $savesub = $true
                            }
                            elseif ($clipframein -ge $currentsubarr[$i].TimeStart -and $clipframeout -le $currentsubarr[$i].TimeEnd)
                            {
                                # subtitle is larger than the clip, so the clip is completely within subtitle
                                # SI > C < SO

                                $tmpint1 = $currentsubarr[$i].TimeStart
                                $tmpint2 = $currentsubarr[$i].TimeEnd

                                # in this situation we want to save the sub but warn the user about it, this is likely a heavily edited section where the user will want to take manual intervention to combine or delete multiple overlapping subtitles
                                $tmpstr = ConvertFramesToSrtTime(($clipstart+($currentsubarr[$i].TimeStart - $clipin)))
                                Write-Output "WARNING: Possible overlap sub at $tmpstr, please search the output file for the BESTEDIT-WARNING-OVERLAP warning message(s)"
                                $subwarning = "BESTEDIT-WARNING-OVERLAP>>>"
                                $savesub = $true
                            }
                            elseif ($clipframein -gt $currentsubarr[$i].TimeStart -and $clipframein -lt $currentsubarr[$i].TimeEnd -and $clipframeout -gt $currentsubarr[$i].TimeEnd)
                            {
                                # subtitle overlaps the start of the clip partially
                                # SI < CI  SO < CO  

                                # need to calculate what percentage of the subtitle is overlapping the video to calculate if it's worth keeping
                                $tmplen = $currentsubarr[$i].TimeEnd - $currentsubarr[$i].TimeStart
                                $tmpoverlap = $currentsubarr[$i].TimeEnd - $clipframein
                                $tmppct = $tmpoverlap / $tmplen

                                if ($tmppct -ge $shortenthreshold)
                                {
                                    # within the threshold amount (70% or above by default), need to shrink the sub start/end to match the audio cuts
                                    $thresholdstart = $tmplen - $tmpoverlap
                                }

                                if ($tmppct -gt 0.50)
                                {
                                    $tmpint1 = $currentsubarr[$i].TimeStart
                                    $tmpint2 = $currentsubarr[$i].TimeEnd

                                    $tmpstr = ConvertFramesToSrtTime(($clipstart+($currentsubarr[$i].TimeStart - $clipin)))
                                    Write-Output "Notice: Partial overlap pct:'$tmppct' start, sub at $tmpstr, please search the output file for the relevant timecode"
                                    #$subwarning = "BESTEDIT-WARNING-OVERLAP2>>>"
                                    $savesub = $true
                                }
                            }
                            elseif ($clipframeout -lt $currentsubarr[$i].TimeEnd -and $clipframeout -gt $currentsubarr[$i].TimeStart -and $clipframein -lt $currentsubarr[$i].TimeStart)
                            {
                                # subtitle overlaps the end of the clip partially
                                # CI < SI  CO < SO 
                                
                                # need to calculate what percentage of the subtitle is overlapping the video to calculate if it's worth keeping
                                $tmplen = $currentsubarr[$i].TimeEnd - $currentsubarr[$i].TimeStart
                                $tmpoverlap = $clipframeout - $currentsubarr[$i].TimeStart
                                $tmppct = $tmpoverlap / $tmplen

                                if ($tmppct -ge $shortenthreshold)
                                {
                                    # within the threshold amount (70% or above by default), need to shrink the sub start/end to match the audio cuts
                                    $thresholdend = $tmplen - $tmpoverlap
                                }

                                if ($tmppct -gt 0.50)
                                {
                                    $tmpint1 = $currentsubarr[$i].TimeStart
                                    $tmpint2 = $currentsubarr[$i].TimeEnd

                                    $tmpstr = ConvertFramesToSrtTime(($clipstart+($currentsubarr[$i].TimeStart - $clipin)))
                                    Write-Output "Notice: Partial overlap pct:'$tmppct' end, sub at $tmpstr, please search the output file for the relevant timecode"
                                    #$subwarning = "BESTEDIT-WARNING-OVERLAP3>>>"
                                    $savesub = $true
                                }
                            }

                            if ($savesub -eq $true)
                            {
                                $tmpstr = $currentsubarr[$i].Subtext
                                $tmpint1 = $currentsubarr[$i].TimeStart
                                $tmpint2 = $currentsubarr[$i].TimeEnd

                                #$output_arr += $currentsubarr[$i] 
                                # we can't just assign it though because the TimeStart and TimeEnd are source video times, we need to shift the subtitle to match our edit time
                                # first find offset from the start of the clip
                                [decimal]$tmpdiff = $currentsubarr[$i].TimeStart - $clipin

                                # then add this offset to the position of the clip
                                $newsub = @{}
                                $newsub.TimeStart = $clipstart + $thresholdstart + $tmpdiff

                                $tmpdiff = $currentsubarr[$i].TimeEnd - $clipin
                                $newsub.TimeEnd = $clipstart - $thresholdend + $tmpdiff

                                $tmpint1 = $newsub.TimeStart
                                $tmpint2 = $newsub.TimeEnd

                                $tmpstr = ConvertFramesToSrtTime($tmpint1)

                                $newsub.Subtext = $subwarning + $currentsubarr[$i].Subtext
                                $output_arr += $newsub
                            }
                        }
                    }

                    # clear values so we don't reprocess same data
                    $masterclip_id = -1
                    $clipspeed = -1
                    $clipframein = 0
                    $clipframeout = 0
                    $clipstart = -1
                    $clipend = -1
                    $clipin = -1
                    $clipout = -1
                }

            }

        }

        # sort array by TimeStart
        $output_arr = ($output_arr | Sort-Object -Property {$_.TimeStart}) 

        # write array out
        $outstream = New-Object IO.StreamWriter ($outputpath + $origname1NOext + ".srt")

        for ($i=0;$i -lt $output_arr.Count;$i++)
        {
            [string]$tmpstr = ($i+1)
            $tmpstr += "`n"

            $tmpstr += ConvertFramesToSrtTime($output_arr[$i].TimeStart)
            $tmpstr += " --> "
            $tmpstr += ConvertFramesToSrtTime($output_arr[$i].TimeEnd)
            $tmpstr += "`n"

            $tmpstr += $output_arr[$i].Subtext
            $tmpstr += "`n" # only need one newline since WriteLine adds another

            $outstream.WriteLine($tmpstr)
        }

        $outstream.Close()

    }

}
else
{
    Write-Output "No matching subtitle files found, aborting processing!"
}